#!/bin/bash

# Install-mod: Convert and install mods for Baldur's Gate to the Mac.
#              Based on TuMac by Loriel

shopt -s extglob
ulimit -n 2048
ulimit -d unlimited
export LANG=C

uncased() {	# make a case-insensitive pattern out of a string
    local y="$1" x z=
    while [[ $y ]]; do
	x=${y%%\$\{*}
	y=${y:${#x}}
	x=${x//[Aa]/[Aa]} x=${x//[Bb]/[Bb]} x=${x//[Cc]/[Cc]} x=${x//[Dd]/[Dd]}
	x=${x//[Ee]/[Ee]} x=${x//[Ff]/[Ff]} x=${x//[Gg]/[Gg]} x=${x//[Hh]/[Hh]}
	x=${x//[Ii]/[Ii]} x=${x//[Jj]/[Jj]} x=${x//[Kk]/[Kk]} x=${x//[Ll]/[Ll]}
	x=${x//[Mm]/[Mm]} x=${x//[Nn]/[Nn]} x=${x//[Oo]/[Oo]} x=${x//[Pp]/[Pp]}
	x=${x//[Qq]/[Qq]} x=${x//[Rr]/[Rr]} x=${x//[Ss]/[Ss]} x=${x//[Tt]/[Tt]}
	x=${x//[Uu]/[Uu]} x=${x//[Vv]/[Vv]} x=${x//[Ww]/[Ww]} x=${x//[Xx]/[Xx]}
	x=${x//[Yy]/[Yy]} x=${x//[Zz]/[Zz]}
	if [[ $2 ]]; then
	    x=${x//\ /\\\ } x=${x//\*/\\*} x=${x//\?/\\?} x=${x//\'/\\\'}
	fi
	z=$z$x
	if [[ $y ]]; then
	    x=${y%%\}*}
	    z=$z$x
	    y=${y:${#x}}
	fi
    done
    echo "$z"
}

tolower() {	# translate upper to lower
    local x="$1"
    x=${x//A/a} x=${x//B/b} x=${x//C/c} x=${x//D/d} x=${x//E/e} x=${x//F/f}
    x=${x//G/g} x=${x//H/h} x=${x//I/i} x=${x//J/j} x=${x//K/k} x=${x//L/l}
    x=${x//M/m} x=${x//N/n} x=${x//O/o} x=${x//P/p} x=${x//Q/q} x=${x//R/r}
    x=${x//S/s} x=${x//T/t} x=${x//U/u} x=${x//V/v} x=${x//W/w} x=${x//X/x}
    x=${x//Y/y} x=${x//Z/z}
    echo "$x"
}

create_bash_script() {
    local x y name=${1//[^a-zA-Z_0-9]/_}
    echo '#!/bin/bash'
    echo 'shopt -s nocaseglob'
    echo 'ulimit -d unlimited'
    echo 'if [[ ! $Arg ]]; then Arg=("$0" "$@") Exit="rm -f /tmp/$$-*"; else Exit="tee /tmp/$$-exit"; fi'
    echo 'if [[ $Arg != /* ]]; then Arg=$PWD/${Arg//\/.\///}; fi'
    echo ''
    echo "Start_$name () {"

    # define the variables for remembering context in select_batch_command
    local cdpat=	# keep track of current directory
    local deferred=	# for defining dynamically scoped variables here
    local redir=("")	# context for nested compound statements
    local save_til_end=()	# label/routine defined in compound statement
    local start_saving=(0)	# start save_til_end
    local OUTPUT	# output of select_batch_command; can't do var=$(...)
    local last_com_was_call=	# return result from function
    local sub_bat=1	# output '(source xxx)' instead of 'bash xxx'
    local header_stuff	# move comments in front of label (mostly)
    unset header_stuff	# so ${#header_stuff[*]} starts as 0

    while read -r; do
	x=${REPLY%%[^$' \t']*}
	if [[ ${REPLY:${#x}} == \:[A-Za-z0-9]* ]]; then    # found :label
	    REPLY=`tolower "${REPLY/:/_}"`
	    if [[ ${#redir[*]} -le 1 ]]; then	# not inside compound statement
		if [[ ${#header_stuff[*]} -ne 0 ]]; then
		    echo -n $'}\n\n'"$header_stuff"
		    unset header_stuff
		else
		    echo "  $REPLY"$'\n}\n\n'"${save_til_end[*]}"
		    save_til_end=()
		fi
		echo "${REPLY:${#x}} () {"
	    elif [[ $start_saving -eq ${#redir[*]} ]]; then
		save_til_end=$save_til_end"  $REPLY"$'\n}\n\n'"${REPLY:${#x}} () {"$'\n'
	    else
		if [[ $start_saving -ne 0 ]]; then
		    save_til_end=$save_til_end"  $REPLY"$'\n}\n\n'
		    save_til_end=("" "${save_til_end[@]}")
		    start_saving=(0 "${start_saving[@]}")
		else
		    echo "  $REPLY"
		fi
		save_til_end=$save_til_end"${REPLY:${#x}} () {"$'\n'
		start_saving=${#redir[*]}
	    fi
	else
	    if [[ $start_saving -eq 0 ]]; then
		select_batch_command "$REPLY"
		if [[ ${#header_stuff[*]} -eq 0 ]]; then
		    echo "  $OUTPUT"
		    if [[ ${#redir[*]} -le 1 && ( ${REPLY:${#x}} == [Ee][Xx][Ii][Tt]* || ${REPLY:${#x}} == [Gg][Oo][Tt][Oo]* ) ]]; then
			header_stuff=	# start next label header
		    fi
		elif [[ $OUTPUT != *[^$' \t']* && ! $header_stuff ]]; then
		    echo ""
		elif [[ ${REPLY%%:*} == *[^$' \t']* ]]; then
		    sed 's/^/  /' <<<"$header_stuff"
		    echo "  $OUTPUT"
		    unset header_stuff
		else
		    header_stuff=$header_stuff$OUTPUT$'\n'
		fi
	    else
		select_batch_command "$REPLY"
		if [[ $start_saving -eq 0 ]]; then
		    echo "  $OUTPUT"
		else
		    save_til_end=$save_til_end"  $OUTPUT"$'\n'
		fi
	    fi
	    if [[ $deferred ]]; then
		eval "$deferred"
		deferred=
	    fi
	fi
    done

    if [[ ${#header_stuff[*]} -ne 0 ]]; then
	echo -n $'}\n\n'"$header_stuff"
    else
	echo $'  exit `$Exit <<<${?}`\n'"${save_til_end[@]}"$'}\n'
    fi
    echo "Start_$name"
}

modify_variable () {		# interpret %~Xv or %%~Xv type of modifying
    case "$2" in
      "")	echo \$\{$1//\\\"\};;
      [Nn][Xx]) echo \$\{$1##\*/\};;
      [Dd][Pp]) echo \$\{$1%/\*\};;
      [Pp])	echo \$\{$1%/\*\};;
      [Dd])	echo /;;
      [Xx])	echo .\$\{$1##\*.\};;
      [Ff])	echo \$\{$1\${PWD};;
#      [Nn])	echo '`v=${'"$1"'##*/};cat<<<${v%.*}`';;
      [Nn])	echo '`v=${'"$1"'##*/};{echo,${v%.*}}`';;
      *)	echo \$\{$1\};;
    esac
}

variable_replace() {		# replace %var% (or !var!) with value
    local x= y= var v z w q

    #echo "IN: '$1'  (redir*=${redir[*]})" 1>&2
    x=${1//^&/\\&}
    #x=${x//^}		# BAT escape character
    x=${x//$'\x03'}	# \x03 used for marking strings
    #if [[ ${redir[*]} == *\ done* ]]; then	# inside a FOR loop
	#exc="!"
    #fi
    while [[ $x == *[%!]* ]]; do
	v=${x%%[%!]*}
	y=$y$v
	x=${x:${#v}}
	v=${x:0:1}	# either % or !
	x=${x:1}

	if [[ $v == % && $x == [\*0-9]* ]]; then  # %n parameter -> ${Arg[n]}
	    y=$y\$\{Arg\[${x:0:1}\]\}
	    x=${x:1}
	elif [[ $v == % && $x == \~* ]]; then
	    x=${x:1}
	    z=${x%%[0-9]*}
	    x=${x:${#z}}
	    y=$y`modify_variable Arg\[${x:0:1}\] "$z"`
	    x=${x:1}
	elif [[ $v == % && $x == %* ]]; then	# FOR loop variable %%x (1 char)
	    x=${x:1}				# becomes shell var ${_x_}
	    if [[ $x == \~* ]]; then		# modify variable
		z=${x:1}
		z=${z%%[^a-zA-Z]*}
		x=${x:${#z}+1}
		y=$y`modify_variable _${z: -1}_ ${z:0:${#z}-1}`
	    else
		y=$y\$\{_${x:0:1}_\}
		x=${x:1}
	    fi
	elif [[ $x == +([A-Za-z0-9_.])?(:*([^$v$' \t']))$v* ]]; then  # %var%
	    var=${x%%[:$v]*}
	    x=${x:${#var}}
	    if [[ $x != \:* ]]; then
		x=${x:1}
		w=
	    else
		w=$x x=
		while [[ $w ]]; do
		    if [[ $w == *%*%* ]]; then
			z=\%
		    elif [[ $w == *!*!* ]]; then
			z=\!
		    else
			break
		    fi
		    q=$w
		    w=${w#*$z}
		    x=$x${q:0:${#q}-${#w}-1}
		    q=${w%%$z*}
		    x=$x\$_$q
		    w=${w:${#q}+1}
		done
		x=$x$w

		if [[ $x == \:\~* ]]; then	# substring: %var:~nnn{,nnn}%
		    w=${x%%$v*}
		    x=${x:${#w}+1}
		    w=\:${w:2}
		    w=${w/,/:}
		else				# string replace: %var:s1=s2%
		    w=//
		    if [[ $x == \:\** ]]; then	# replace at the beginning
			x=${x:1}
			w=/\#
		    fi
		    z=${x%%=*}
		    x=${x:${#z}+1}
		    z=${z//\\/\\\\}
		    z=${z//\//\\/}
		    w=$w${z:1}
		    z=${x%%$v}
		    x=${x:${#z}+1}
		    w=$w/${z//\\/\\\\}
		    w=${w//\"/\\\"}
		    w=${w//\'/\`echo -e \\\\x27\`}   # because \' doesn't work
	    fi  fi

	    if [[ $var == [Ee][Rr][Rr][Oo][Rr][Ll][Ee][Vv][Ee][Ll] ]]; then
		if [[ $last_com_was_call ]]; then
		    y=$y'`{cat,/tmp/${$}-exit}`'
		else
		    y=$y'${?}'
		fi
	    elif [[ $var == [Cc][Dd] ]]; then
		if [[ $y == *\" ]]; then
		    y=$y'${PWD}'
		else
		    y=$y'"${PWD}"'
		fi
	    else
		local myname=`tolower _${var//./_}`
		z=${myname}__set
		#echo -n repvar: 1>&2
		#if [[ $2 ]]; then echo -n " 2=$2" 1>&2; fi
		#if [[ $v == \! ]]; then echo -n " v=!" 1>&2; fi
		#if [[ ${!z} != 1 ]]; then echo -n " !z=${!z}" 1>&2; fi
		#if [[ $y == *[Ss][Ee][Tt]+([$' \t/Aa'])$var[$'= \t']* ]]; then echo -n " set" 1>&2; fi
		#if [[ ${!myname%% *} == *\$\{* ]]; then echo -n " myname=${!myname%% *}" 1>&2; fi
		#echo 1>&2
		if [[ $2 || $v == \! || ${!z} != 1 || $y == *[Ss][Ee][Tt]+([$' \t/Aa'])$var[$'= \t']* || ${!myname%% *} == *\$\{* ]]; then
		    y=$y\$\{$myname$w\}	    # output "${_var}"
		else			    # or its value if SET exactly once
		    y=$y$(eval echo \"\${$myname$w}\")
	    fi	fi
	else
	    y=$y$v
	fi
    done

    echo "$y$x"
    #echo "OUT: $y$x" 1>&2
}

dequote () {
    local B C D= x=\[$' \t'\]\*

    #echo "DEQUOTE: '$1'" 1>&2
    C=${1%%[^$' \t']*}
    B=${1:${#C}}
    if [[ $B == \#* ]]; then
	D=\\
    fi
    while [[ $B ]]; do
	C=${B%%[$'"$ \t&()[]\'']*}
	B=${B:${#C}}
	D=$D${C//\\//}

	if [[ ! $B || $B == $x ]]; then
	    C=${B%%[^$' \t']*}
	    B=${B:${#C}}
	    if [[ ! $2 ]]; then
		break
	    fi
	    D=${D/%\*.\*/*}" "
	    if [[ $B == \#* ]]; then
		D=$D\\
	    fi
	elif [[ $B == \"* ]]; then
	    B=${B:1}
	    if [[ ! $x ]]; then
		x=\[$' \t'\]\*
	    elif [[ $B == \"* ]]; then
		B=${B:1}
		D=$D\"\"
	    else
		x=
	    fi
	elif [[ $B == \$\{* ]]; then
	    C=${B%%\}*}
	    B=${B:${#C}+1}
	    #if [[ $C == \$\{*([A-Za-z0-9_]) ]]; then
		#D=$D\"${C/{}\"
	    #else
		#D=$D\"$C\}\"
	    #fi
	    if [[ ! $x || $C == '${Arg['* ]]; then
		D=$D\"$C\}\"
	    else
		D=$D$C\}
	    fi
	elif [[ $D == *\\ && $D != *\\\\ ]]; then
	    D=$D${B:0:1}
	    B=${B:1}
	else
	    D=$D\\${B:0:1}
	    B=${B:1}
	fi
    done

    D=${D/%\*.\*/*}
    #echo "ENQUOTE: '$D'  remaining(${#B}): '$B'" 1>&2
    if [[ $2 ]]; then
	echo "$D"
    else
	echo "$D"$'\x03'"$B"
    fi
}

select_batch_command() {	# Args: "line" dont_interpolate/dont_replace
    local A B options com not x y z ind tokens
    local cnv=("==" "<=" "<" ">=" ">" "!=")
    local cne=("-eq" "-le" "-lt" "-ge" "-gt" "-ne")
    local out=${1%%[^$' \t']*}	# indentation
    local line=${1:${#out}}

    if [[ ! $line ]]; then
	OUTPUT=$out
	return
    fi
    if [[ $line == "::"* ]]; then
	OUTPUT="$out# ${line:2}"
	return
    fi

    line=${line#@}
    if [[ $2 != dont_replace ]]; then
	line=$(variable_replace "$line" $2)
    fi

    if [[ $line == ":"* ]]; then
	OUTPUT="$out# ${line:1}"
	return
    fi

    last_com_was_call=		# reset using /tmp/$$-exit instead of ${?}

    while [[ $line ]]; do	# special processing for compound statements
	case $line in
	  [Ii][Ff][$' \t']*)		# IF statement
	    if [[ $redir == "; fi"* ]]; then
		redir=${redir#????}
	    else
		out=$out"if [[ "
	    fi

	    line=${line#??}
	    options=$line
	    while [[ $line ]]; do
		x=${line%%[^$' \t']*}
		line=${line:${#x}}
		if [[ $line != /* ]]; then
		    break
		fi
		line=${line#/*[$' \t']}
	    done
	    options=${options:0:${#options}-${#line}}

	    not=
	    if [[ $line == [Nn][Oo][Tt][$' \t']* ]]; then
		not="! "
		x=${line#???*[^$' \t']}
		line=${line:${#line}-${#x}-1}
	    fi
	    if [[ $line == [Ee][Xx][Ii][Ss][Tt][$' \t']* ]]; then
		line=$(dequote "${line#?????}")
		x=${line%$'\x03'*}
		line=${line:${#x}+1}
		if [[ $x == */[Nn][Uu][Ll] ]]; then
		    out=$out"$not-d ${x/%\/[Nn][Uu][Ll]//.}"
		elif [[ ${x//\`*([^\`])\`} != *[?*]* ]]; then
		    out=$out"$not-e $x"
		else
		    out=$out"$not\$(ls -Ld $x 2>/dev/null)"
		fi
	    else
		y=${line%%==*}
		x=$(dequote "$y")
		A=${x%$'\x03'*}
		line=${line:${#y}-${#x}+${#A}+1}

		let ind=0
		if [[ $line == "=="* ]]; then
		    line=${line:2}
		else
		    for x in "[Ee][Qq][Uu]" "[Ll][Ee][Qq]" "[Ll][Ss][Ss]" "[Gg][Ee][Qq]" "[Gg][Tt][Rr]" "[Nn][Ee][Qq]"; do
			if [[ $line == $x* ]]; then
			    line=${line:3}
			    break
			fi
			if (( $ind >= 5 )); then
			    break
			fi
			let ind++
		    done
		fi
		if [[ $not ]]; then
		    let ind=5-$ind
		fi

		line=$(dequote "$line")
		B=${line%$'\x03'*}
		line=${line:${#B}+1}

		if [[ $options == */[iI]* ]]; then
		    if [[ $B != *\$\{* ]]; then
			B=$(uncased $B)
		    elif [[ $A != *\$\{* ]]; then
			x=$A
			A=$B
			B=$(uncased $x)
		    else
			A='$(tr a-z A-Z <<<'"\"$A\")"
			B='$(tr a-z A-Z <<<'"\"$B\")"
		fi  fi

		if [[ $A == '${?}' || $A == '$?' ]]; then
		    out=$out"$A ${cne[$ind]} \$(($B % 256))"
		elif [[ $A == '`{cat,/tmp/'*'-exit}`' ]]; then
		    out=$out"$A ${cne[$ind]} $B"
		else
		    out=$out"$A ${cnv[$ind]} $B"
	    fi  fi

	    redir="; fi$redir"
	    if [[ $line == [Ii][Ff][$' \t']* ]]; then
		out=$out" && "
	    else
		out=$out" ]]; then "
	    fi;;

	  \(*)			# begin compound statement
	    redir=("" "${redir[@]}")
	    line=${line:1}
	    x=${line%%[^$' \t']*}
	    if [[ ${#x} == ${#line} ]]; then
		line=
		continue
	    fi
	    line=${line:${#x}}
	    if [[ $line != *\)* ]]; then
		continue
	    fi

	    x=${line%%)*}
	    select_batch_command "$x" dont_replace
	    out=$out${OUTPUT/##/:\ }
	    line=${line:${#x}}
	    ;;

	  \)*)			# end compound statement
	    if [[ $start_saving -ge ${#redir[*]} ]]; then
		save_til_end=$save_til_end$'}\n\n' 
		unset start_saving[0]
		start_saving=("${start_saving[@]}")
	    fi
	    line=${line:1}
	    x=${line%%[^$' \t']*}
	    if [[ ${#x} == ${#line} ]]; then
		line=
	    else
		line=${line:${#x}}
	    fi
	    unset redir[0]
	    redir=("${redir[@]}")

	    if [[ $line != [Ee][Ll][Ss][Ee][$' \t']* ]]; then
		x=${1#*[^$' \t']}
		if [[ ${1:${#1}-${#x}-1:1} == \) ]]; then
		    redir=${redir#\;\ }
		fi
		out=$out$redir
		redir=
		continue
	    fi
	    x=${line#????*[^$' \t']}
	    line=${line:${#line}-${#x}-1}
	    if [[ $line == [Ii][Ff][$' \t']* ]]; then
		out=$out"; elif [[ "
		redir="; if$redir"
		continue
	    fi
	    x=${1#*[^$' \t']}
	    if [[ ${1:${#1}-${#x}-1:1} != \) ]]; then
		out=$out"; "
	    fi
	    out=$out"else "
	    ;;

	  [Ff][Oo][Rr][$' \t']*)	# FOR statement
	    if [[ $line != *\)* ]]; then
		while read -r; do
		    x=${REPLY%%[^$' \t']*}
		    x=${REPLY:${#x}}
		    if [[ $x && $x != \:* ]]; then
			line=$line\ $(variable_replace "$x")
			if [[ $x == *\)* ]]; then
			    break
		    fi  fi
		done
	    fi
	    x=${line#???*[^$' \t']}
	    line=${line: -1-${#x}}
	    options=${line:0:2}
	    x=${line#/?*[^$' \t']}
	    if [[ ${#x} != ${#line} ]]; then
		line=${line: -1-${#x}}
	    fi
	    local parsing=
	    if [[ $line != \$\{_?_\}* ]]; then
		if [[ $line == \"* ]]; then
		    parsing=${line:1}
		    parsing=${parsing%%\"*}
		    line=${line:${#parsing}+2}
		else
		    parsing=${line%%[$' \t']*}
		    line=${line:${#parsing}}
		fi
		x=${line%%[^$' \t']*}
		line=${line:${#x}}
	    fi
	    A=${line%%[$' \t']*}
	    line=${line:${#A}}
	    A=${A:2:3}
	    x=${line#*\(*[^$' \t']}
	    line=${line: -1-${#x}}
	    B=${line%%\)*}
	    line=${line:${#B}+1}
	    x=${line#*[$' \t'][Dd][Oo][$' \t']*[^$' \t']}
	    line=${line: -1-${#x}}

	    case "$options" in
	      /[Dd])
		out=$out"for $A in "
		y=()
		while [[ $B ]]; do
		    B=$(dequote "$B")
		    x=${B%$'\x03'*}
		    B=${B:${#x}+1}
		    if [[ $x == *[*?]* ]]; then
			y[${#y[*]}]=$x/.
		    else
			y[${#y[*]}]=$x
		    fi
		done
		out=$out"${y[@]}; do if [[ \$$A == *[?*]* ]]; then continue; fi; $A=\${$A%/.}; "
		;;
	      /[Rr])
		y=
		if [[ $B == "." ]]; then
		    B=
		else
		    while [[ $B ]]; do
			B=$(dequote "$B")
			x=${B%$'\x03'*}
			B=${B:${#x}+1}
			y=$y"${y:+ -o }-iname \"$x\""
		    done
		fi
		if [[ $parsing ]]; then
		    x=$(dequote "$parsing" all)
		else
		    x="*"
		fi
		out=$out"IFS=\$'\n'; for $A in \$(find -H $x \\( $y \\) -print); do IFS=\$' \t\n'; ";;
	      /[Ff])
		local eol= delims= tokens=() files= skip=
		if [[ $parsing == *[Ee][Oo][Ll]=* ]]; then
		    x=${parsing#*[Ee][Oo][Ll]=}
		    eol=" -d \"${x:0:1}\""
		fi
		if [[ $parsing == *[Ss][Kk][Ii][Pp]=* ]]; then
		    x=${parsing#*[Ss][Kk][Ii][Pp]=}
		    skip=" | tail -"${x%%[$' \t']*}
		fi
		if [[ $parsing == *[Dd][Ee][Ll][Ii][Mm][Ss]=* ]]; then
		    x=${parsing#*[Dd][Ee][Ll][Ii][Mm][Ss]=}
		    delims=" IFS=\"${x#*[Dd][Ee][Ll][Ii][Mm][Ss]=}\""
		fi
		if [[ $parsing == *[Tt][Oo][Kk][Ee][Nn][Ss]=* ]]; then
		    x=${parsing#*[Tt][Oo][Kk][Ee][Nn][Ss]=}
		    x=${x%%\ *}
		    y=${x##*[-,]}
		    if [[ $y == *\* ]]; then
			y=${y%\*}
			x="${x:0:${#x}-1},$(($y+1))"
		    fi

		    tokens=(`eval echo {0..$(($y+1))}`)
		    tokens=("${tokens[@]/#*/junk}")
		    unset tokens[0]

		    if [[ $A == _[a-z]_ ]]; then z=z; else z=Z; fi
		    local q=(`eval echo "{${A:1:1}..$z}"`)
		    for x in ${x//,/\ }; do
			if [[ $x == *-* ]]; then
			    x=`eval echo "{${x/-/..}}"`
			fi
			while [[ $x ]]; do
			    y=${x%%\ *}
			    tokens[$y]=_${q}_
			    unset q[0]
			    q=("${q[@]}")
			    x=${x:${#y}+1}
			done
		    done
		else
		    tokens=($A junk)
		fi

		y=${B:0:1}
		if [[ $parsing == *[Uu][Ss][Ee][Bb][Aa][Cc][Kk][Qq]* ]]; then
		    if [[ $y == \' ]]; then
			y=\"
		    elif [[ $y == \` ]]; then
			y=\'
		    else
			y=
		fi  fi

		if [[ $y == \" ]]; then
		    redir=" <<<$B"$redir
		else
		    redir=' </tmp/$$-temp$U'$redir
		    if [[ $y == \' ]]; then
			redir=("" "${redir[@]}")
			select_batch_command "${B//${B:0:1}}" dont_replace
			unset redir[0]
			redir=("${redir[@]}")
			out=$out"let U++; ${OUTPUT/##/: }"' >/tmp/$$-temp$U; '
		    else
			out=$out"let U++; sed s/$'\r$'// "$(dequote "$B" all)' >/tmp/$$-temp$U; '
		fi  fi

		out=$out"while$delims read -r$eol ${tokens[@]}; do"
		for y in "${tokens[@]}"; do
		    if [[ $y != junk ]]; then
			out=$out" $y=\${$y//\\\\//}"
		    fi
		done
		out=$out"; "
		unset q tokens;;

	      *)
		out=$out"for $A in "$(dequote "$B" all)"; do "
		if [[ $B == *[?*]* ]]; then
		    out=$out"if [[ \$$A == *[?*]* ]]; then continue; fi; "
		fi;;
	    esac

	    redir="; done$redir"
	    ;;

	  *) break;;
	esac
    done

    z=$line line=
    while [[ $z ]]; do		# & - command separator?
	x=${z#*[\\\&\$\"]}
	if [[ ${#x} -eq ${#z} ]]; then
	    line=$line$z
	    break
	fi
	line=$line${z:0:${#z}-${#x}}
	z=$x
	if [[ $line == *\\ ]]; then
	    line=$line${z:0:1}
	    z=${z:1}
	elif [[ $line == *\& ]]; then
	    if [[ $line == *[\\\>\<]\& ]]; then
		continue
	    fi
	    redir=("" "${redir[@]}")
	    select_batch_command "${line%&}" dont_replace
	    unset redir[0]
	    redir=("${redir[@]}")
	    out=$out"${OUTPUT/##/: }; "
	    line=
	    x=${z%%[^$' \t']*}
	    z=${z:${#x}}
	elif [[ $line == *\" ]]; then
	    x=${z#*\"}
	    if [[ ${#x} -eq ${#z} ]]; then
		line=$line$z
		break
	    fi
	    line=$line${z:0:${#z}-${#x}}
	    z=${z:${#z}-${#x}}
	elif [[ $z == \{* ]]; then
	    x=${z%%\}*}
	    line=$line$x\}
	    z=${z:${#x}+1}
	fi
    done

    if [[ $line == [Ss][Ee][Tt][$' \t']* ]]; then  # SET needs to be done early
	x=${line#???*[^$' \t']}
	line=${line:${#line}-${#x}-1}
	options=
	if [[ $line == /* ]]; then
	    options=${line:1:1}
	    x=${line#??*[^$' \t']}
	    line=${line:${#line}-${#x}-1}
	fi
	if [[ $line == \"* ]]; then
	    line=${line:1}
	    line=${line%\"*}
	fi
	var=${line%%[$' \t'=]*}
	var=`tolower _${var//./_}`

	y=${line#*=}
	local raw_line=$y
	line=
	while [[ $y ]]; do
	    x=${y%%\$\{*}
	    y=${y:${#x}+2}
	    line=$line${x//\\//}
	    if [[ $y ]]; then
		x=${y%%\}*}
		y=${y:${#x}}
		line=$line\$\{$x
	    fi
	done

	varset=${var}__set

	if [[ ${!varset} ]]; then
	    eval let $varset++
	elif [[ $deferred == *" $varset=1"* || ${redir[*]} == *\ done* ]]; then
	    deferred=$deferred"local $varset=2;"
	else
	    deferred=$deferred"local $varset=1;"
	fi

	if [[ $options == [aA] ]]; then
	    if [[ ${!varset} ]]; then
		eval "$var=\$(($line))"
	    elif [[ $deferred == *" $var="* ]]; then
		deferred=$deferred"$var=\$(($line));"
	    else
		deferred=$deferred"local $var=\$(($line));"
	    fi
	    out=$out"$var=\$(($line))"
	elif [[ $options == [Pp] ]]; then
	    deferred=$deferred"local $var $varset=2;"
	    out=$out"read -p '${line//'/'\\''}' $var"
	elif [[ $line == *\$\{* || $line == *\`*\`* ]]; then
	    x=
	    while [[ $line ]]; do
		y=${line%%[\"\$]*}
		line=${line:${#y}}
		x=$x$y
		if [[ $line == \"* ]]; then
		    x=$x\\\"
		    line=${line:1}
		elif [[ $line == \$\{* ]]; then
		    y=${line%%\}*}\}
		    x=$x$y
		    line=${line:${#y}}
		else
		    x=$x${line:0:1}
		    line=${line:1}
		fi
	    done
	    if [[ ${!varset} ]]; then
		eval "$var='\"$x\"'"
	    elif [[ $deferred == *" $var="* ]]; then
		deferred=$deferred"$var='\"$x\"';"
	    else
		deferred=$deferred"local $var='\"$x\"';"
	    fi
	    deferred=$deferred"$varset=2;"
	    out=$out"$var=\"$x\""
	else
	    x=${line//\'/\'\\\'\'}
	    if [[ ${!varset} ]]; then
		eval "$var='$x'"
	    elif [[ $deferred == *" $var="* ]]; then
		deferred=$deferred"$var='$x';"
	    else
		deferred=$deferred"local $var='"${raw_line//\'/\'\\\'\'}"';"
	    fi
	    out=$out"$var='$x'"
	fi

	OUTPUT="$out$redir"
	redir=
	return
    fi

    while [[ $line == *\|* ]]; do	# output redirected to program
	x=${line%%|*}
	line=${line##*([^\|])\|*([$' \t'])}
	redir=("" "${redir[@]}")
	select_batch_command "$x" dont_replace
	unset redir[0]
	redir=("${redir[@]}")
	out=$out"${OUTPUT/##/: } | "
    done

    while [[ $line == *[\<\>]* ]]; do	# output redirected to file
	x=$line
	line=${line%[\<\>]*}
	if [[ $line == *\< ]]; then
	    line=$x
	    break
	fi
	y=${x:${#line}:1}
	x=${x:${#line}+1}
	if [[ $x == \&* ]]; then
	    if [[ $x == \&[0-9]* ]]; then
		y=$y${x:0:2}; x=${x:2}
	    else
		y=$y${x:0:1}; x=${x:1}
	    fi
	    if [[ $line == *[0-9] ]]; then
		y=${line: -1}$y
		line=${line:0:${#line}-1}
	    fi
	elif [[ $y == \> ]]; then
	    if [[ $line == *\> ]]; then
		y=\>\>
		line=${line:0:${#line}-1}
	    fi
	    if [[ $line == *[$' \t'][1-9] ]]; then
		y=${line: -1}$y
		line=${line:0:${#line}-1}
	fi  fi
	if [[ $y != *[0-9] ]]; then
	    x=$(dequote "$x")
	    x=${x%$'\x03'*}
	    if [[ $x == ?(*/)[Nn][Uu][Ll] ]]; then
		x=/dev/null
	    fi
	    y=$y$x
	fi
	redir=\ $y$redir
    done

    # simple command line

    if [[ $line == [Ee][Cc][Hh][Oo]* ]]; then
	A=					# don't remove /X switches
	options=${line##*([^$' \t'])}		# remember spacing
    else
	A=/
	options=
	if [[ $cdpat ]]; then			# change .. to $BGdir
	    z=${cdpat//\//\\\\}
	    line=${line//\"$z\\/\"\$\{BGdir\}\\}
	    line=${line//$z\\/\"\$\{BGdir\}\"\\}
    fi	fi

    x=$line line=
    z=\$\"$A
    while [[ $x ]]; do		# get rid of /X switches and change all \ to /
	y=${x%%[$z]*}
	line=$line${y//\\//}
	if [[ ${#x} == ${#y} ]]; then
	    break
	fi
	x=${x:${#y}}
	if [[ $x == \"* ]]; then
	    x=${x:1}
	    line=$line\"
	    if [[ $z == \"* ]]; then z=\$\"$A; else z=\"\$; fi
	elif [[ $x == \$\{* ]]; then
	    y=${x%%\}*}
	    x=${x:${#y}}
	    line=$line$y
	elif [[ $x == /?\:\"* && $line == *[$' \t'] ]]; then
	    options=$options" $x"
	    x=${x#/?\:\"*\"}
	    options=${options:0:${#options}-${#x}}
	elif [[ $x == /?\:* && $line == *[$' \t'] ]]; then
	    y=${x%%[$' \t']*}
	    x=${x:${#y}}
	    options=$options" $y"
	elif [[ $x == /[-+]?* && $line == *[$' \t'] ]]; then
	    y=${x%%[$' \t']*}
	    x=${x:${#y}}
	    options=$options" $y"
	elif [[ $x == /* && $line == *[$' \t'] ]]; then
	    options=$options" ${x:0:2}"
	    x=${x:2}
	else
	    line=$line${x:0:1}
	    x=${x:1}
	fi
    done

    line=$(dequote "$line")	# command
    com=${line%$'\x03'*}
    line=${line:${#com}+1}
    local whole_line=$line
    line=$(dequote "$line")
    A=${line%$'\x03'*}		# first argument
    line=${line:${#A}+1}

    case $com in
	"") ;;

	":"*)
	    if [[ $redir || $out == *"| " ]]; then
		out=$out": '$whole_line'"
	    else
		out=$out"# $whole_line"
	    fi;;

	...*)
	    if [[ $redir || $out == *"| " ]]; then
		out=$out": $com '$whole_line'"
	    else
		out=$out"# $com $whole_line"
	    fi;;

	*.[Bb][Aa][Tt])
	    A=$mod/${com//\\} x=${A%????}
	    if [[ $A -nt $x.sh ]]; then
		sed s/$'\r$'// "$A" | create_bash_script "${x##*/}" > "$x.sh"
	    else
		y=`grep -i "/${A##*/}$" "$mod/.batlist" 2>/dev/null | head -1`
		if [[ ! $y || ! -e $y ]]; then
		    out=$out"if [[ ! -f \"$x.sh\" ]]; then Install-mod.sh -batch \"$A\"; fi; "
		elif [[ $y -nt ${y%???}sh ]]; then
		    x=${y%????}
		    sed s/$'\r$'// "$y" | create_bash_script "${x##*/}" > "${y%???}sh"
	    fi  fi
	    if [[ $sub_bat ]]; then
		out=$out"(Arg=($com "$(dequote "$whole_line" all)"); source ${com%???}sh)"
	    else
		out=$out"bash ${com%???}sh "$(dequote "$whole_line" all)
	    fi;;

	[Aa][Tt][Tt][Rr][Ii][Bb]?(/*))
	    options=$options\ ${com#??????}
	    x=
	    line=$whole_line
	    while [[ $line == [-+]* ]]; do
		if [[ ${line:0:2} == +[Rr] ]]; then
		    x="-w"
		elif [[ ${line:0:2} == -[Rr] ]]; then
		    x="+w"
		fi
		y=${line:2}
		y=${y%%[^$' \t']*}
		line=${line:${#y}+2}
	    done
	    if [[ ! $x ]]; then
		out=$out"# $com $whole_line"
	    else
		A=$(dequote "$line")
		if [[ $options == */[Ss]* ]]; then
		    x="-R $x"
		fi
		out=$out"chmod -f $x ${A%$'\x03'*}"
	    fi;;

	[Mm][Oo][Rr][Ee]?(/*))
	    if [[ ! $A ]]; then
		out=$out"cat"
	    else
		out=$out"sed s/\$'\r'// $A"
	    fi;;

	sh|bash)	# hm, already changed for Macs
	    out=$out"bash -O nocaseglob "$(dequote "$whole_line" all);;
	*.sh)
	    out=$out"bash -O nocaseglob $com "$(dequote "$whole_line" all);;
	rm|cp|mv)
	    if [[ $whole_line == *[*?]* ]]; then
		out=$out"bash -O nocaseglob -c '$com "$(dequote "$whole_line" all)\'
	    else
		out=$out"$com "$(dequote "$whole_line" all)
	    fi;;

	?(*/)[Pp][Aa][Tt][Cc][Hh]?(.[Ee][Xx][Ee]))
	    out=$out'(for x in $(sed -n $'\''s/\r$//\n/^--- /s,\\\\,/,g\nw /dev/stderr\ns/^--- \\(.*\\)\t.*/\\1/p'\''); do if [[ ! -e "$x"-orig ]]; then mv -f "$x"{,-orig}; sed $'\''s/\r$//'\'' "$x"-orig > "$x"; fi 2>/dev/null; done 2>&1 | patch -N '"$whole_line)";;

	[Ss][Oo][Rr][Tt]?(/*))
	    options=$options\ ${com#????}
	    x=-f
	    if [[ $options == */[Rr]* ]]; then
		x=$x" -r"
	    fi
	    if [[ $options == */+* ]]; then
		y=${options##*/+}
		z=${y##*([0-9])}
		x=$x" -k .${y:0:${#y}-${#z}}"
	    fi

	    y=$(dequote "$whole_line")
	    out=$out"sort $x ${y%$'\x03'*}";;

	*/[Ff][Ii][Nn][Dd])
	    if [[ $A == -* ]]; then
		A=". $A"
	    fi
	    out=$out"find $A "$(dequote "$line" all);;

	[Ff][Ii][Nn][Dd]?(/*))
	    options=$options\ ${com##*([^/])}
	    x=
	    if [[ $options == */[Ii]* ]]; then
		x=$x" -i"
	    fi
	    if [[ $options == */[Vv]* ]]; then
		x=$x" -v"
	    fi
	    if [[ $options == */[Nn]* ]]; then
		x=$x" -n"
	    fi
	    if [[ $options == */[Cc]* ]]; then
		x=$x" -c"
	    fi
	    out=$out"fgrep$x $A "$(dequote "$line" all);;

	[Ff][Ii][Nn][Dd][Ss][Tt][Rr]?(/*))
	    options=$options\ ${com##*([^/])}
	    x=
	    if [[ $options == */[Ii]* ]]; then
		x=$x" -i"
	    fi
	    if [[ $options == */[Vv]* ]]; then
		x=$x" -v"
	    fi
	    if [[ $options == */[Nn]* ]]; then
		x=$x" -n"
	    fi
	    if [[ $options == */[Cc]* ]]; then
		y=${options##*/[Cc]:}
		if [[ $y == \"* ]]; then
		    y=${y:1}; y=${y%%\"*}
		else
		    y=${y%%[$' \t']*}
		fi
	    else
		y=${A//\\ /|} A=
	    fi
	    if [[ $options == */[Bb]* ]]; then
		y=\^$y
	    fi
	    if [[ $options == */[Ee]* ]]; then
		y=$y\$
	    fi
	    out=$out"grep$x -e \"$y\" $A "$(dequote "$line" all);;

	*[Oo][Gg][Gg][Dd][Ee][Cc]*)
	    if [[ $A == -o ]]; then
		line=$(dequote "$line")
		B=${line%$'\x03'*}
		line=$(dequote "${line:${#B}+1}")
		A=${line%$'\x03'*}
		out=$out"sox $A $B"
	    else
		while [[ $A ]]; do
		    x=
		    if [[ $A == *.[Cc][Rr][Mm] ]]; then
			x="-t ogg "
		    fi
		    if [[ $A == *[*?]* ]]; then
			out=$out"for file in $A; do sox $x\$file \${file%.*}.wav; done"
		    elif [[ $A == -w ]]; then
			line=$(dequote "$line")
			B=${line%$'\x03'*}
			line=$(dequote "${line:${#B}+1}")
			A=${line%$'\x03'*}
			line=${line:${#A}+1}
			x=
			if [[ $A == *.[Cc][Rr][Mm] ]]; then
			    x="-t ogg "
			fi
			out=$out"sox $x$A $B"
		    else
			out=$out"sox $x$A ${A%.*}.wav"
		    fi

		    A=${line%$'\x03'*}
		    line=${line:${#A}+1}
		done
	    fi;;

	*[Tt][Ii][Ss][Uu][Nn][Pp][Aa][Cc][Kk]*)
	    x=""
	    if [[ $whole_line != *-s* ]]; then
		x="-s "
	    fi
	    out=$out"tisunpack $x$whole_line";;

	*[Tt][Ii][Ss]2[Bb][Gg]2*)
	    out=$out"tis2bg2 -s ${whole_line//-sw}";;

	*[Mm][Oo][Ss][Uu][Nn][Pp][Aa][Cc][Kk]*)
	    x=""
	    if [[ $whole_line != *-s\ *\ -s* ]]; then
		if [[ $whole_line != *-s* ]]; then
		    x="-s -s "; else x="-s "
	    fi  fi
	    out=$out"mosunpack $x$line";;

	[Dd][Ee][Ll][Tt][Rr][Ee][Ee]?(/*))
	    out=$out"rm -rf $A";;

	[Dd][Ee][Ll]|[Ee][Rr][Aa][Ss][Ee])
	    x=
	    if [[ $options == */[Ss]* ]]; then x=r; fi
	    out=$out"rm -f$x $A "$(dequote "$line" all);;

	[Cc][Dd]..|[Cc][Hh][Dd][Ii][Rr]..|[Cc][Dd]\"*|[Cc][Hh][Dd][Ii][Rr]\"*)
	    cdpat=${cdpat%%?(/)..}
	    out=$out"cd ..";;

	[Cc][Dd]|[Cc][Hh][Dd][Ii][Rr])
	    if [[ $A ]]; then
		out=$out"cd $A"
		if [[ $A == \"\$* ]]; then
		    A=${A/#\"\${BGdir}?(\/)/\"}
		fi
		cdpat=${A//+([^\/])/..}
	    else
		out=$out"echo \$PWD"
	    fi;;

	[Dd][Ii][Rr])
	    if [[ $options == */[Ss]* ]]; then
		if [[ $options == */[Bb]* ]]; then
		    if [[ ! $A ]]; then
			A="*"
		    fi
		    if [[ $A != /* ]]; then
			A=\"\$PWD\"/$A
		    fi
		    out=$out"find $A -print"
		else
		    out=$out"ls -laR $A"
		fi
	    else
		if [[ $A == *[?*]* ]]; then
		    if [[ $options == */[Bb]* ]]; then
			out=$out"ls -1d $A"
		    else
			out=$out"ls -lTdL $A | sed 's,[^ ]* *[^ ]* *[^ ]* *[^ ]* *\([^ ]*\) *\([^ ]*\) *\([^ ]*\) *\([^ ]*\) *\([^ ]*\) *\(.*\),\2/\3/\5 \4 AM \1 \6,'"
		    fi
		elif [[ $options == */[Bb]* ]]; then
		    out=$out"ls -1 $A"
		else
		    out=$out"ls -lTL $A | sed 's,[^ ]* *[^ ]* *[^ ]* *[^ ]* *\([^ ]*\) *\([^ ]*\) *\([^ ]*\) *\([^ ]*\) *\([^ ]*\) *\(.*\),\2/\3/\5 \4 AM \1 \6,'"
	    fi  fi;;

	[Rr][Ee][Nn]?(/*)|[Rr][Ee][Nn][Aa][Mm][Ee]?(/*))
	    if [[ $com == */* ]]; then options=$options" /${com#*/}"; fi
	    line=$(dequote "$line")
	    B=${line%$'\x03'*}
	    line=${line:${#B}+1}

	    if [[ $A == *[*?]* && $B == *[*?]* ]]; then
		x="`uncased ${A##*/} no`"
		pat="`sed -e 's|\([^*?.,]*\)[*?]*\([^.,]*\)\([^*?,]*\)[*?]*\([^,]*\),\([^*?.,]*\)[*?]*\([^.]*\)\([^*?]*\)[*?]*\(.*\)|s,.*\1\\\\([^/]*\\\\)\2\\\\\3\\\\(.*\\\\)\4,\5\\\\1\6\7\\\\2\8,|' <<<$x,$B`"
		out=$out"for file in $A; do mv -f \"\$file\" \`sed '$pat' <<<\"\$file\"\`; done"
	    elif [[ ${A##*/} == *\"\$* ]]; then
		if [[ $A == */* ]]; then
		    out=$out"x=$A; mv -f $A \"\${x%/*}\"/$B"
		else
		    out=$out"if [[ $A == */* ]]; then x=$A; mv -f $A \"\${x%/*}\"/$B; else mv -f $A $B; fi"
		fi
	    elif [[ $A == */* ]]; then
		out=$out"mv -f $A ${A%/*}/$B"
	    else
		out=$out"mv -f $A $B"
	    fi;;

	[Cc][Oo][Pp][Yy]?(/*)|[Xx][Cc][Oo][Pp][Yy]?(/*)|[Mm][Oo][Vv][Ee]?(/*))
	    if [[ $com == */* ]]; then options=$options" /${com#*/}"; fi
	    line=$(dequote "$line")	# second argument
	    B=${line%$'\x03'*}
	    line=${line:${#B}+1}

	    if [[ $com == [Mm]* ]]; then	# MOVE
		com="mv -f"
	    else
		com="cp -fR"			# COPY and XCOPY
		if [[ ${A##*/} == *@([Ss][Ee][Tt][Uu][Pp]|[Ww][Ee][Ii][Dd][Uu])*.[Ee][Xx][Ee] ]]; then
		    extra_name=${B##*/}
	    fi  fi

	    if [[ $B == "" ]]; then
		out=$out"$com ${A//,/ } ."
	    elif [[ ${B##*/} == \*.\* || ${A##*/} == `uncased "${B##*/}"` ]]; then
		if [[ $B == */* ]]; then
		    B=${B%/*}/
		else
		    B="."
		fi
		out=$out"$com ${A//,/ } $B"
	    elif [[ $A == *[*?]* && $B == *[*?]* ]]; then
		x="`uncased ${A##*/} no`"
		# Yes, that's right, 52 backslashes.  Shell scripting is fun.
		pat="`sed -e 's|\([^*?.,]*\)[*?]*\([^.,]*\)\([^*?,]*\)[*?]*\([^,]*\),\([^*?.,]*\)[*?]*\([^.]*\)\([^*?]*\)[*?]*\(.*\)|s,.*\1\\\\([^/]*\\\\)\2\\\\\3\\\\(.*\\\\)\4,\5\\\\1\6\7\\\\2\8,|' <<<$x,$B`"
		out=$out"for file in $A; do $com \"\$file\" \`sed '$pat' <<<\"\$file\"\`; done"
	    elif [[ ${A##*/} == *.* && ${A##*.} == `uncased "${B##*.}"` ]]; then
		out=$out"$com ${A//,/ } $B"
	    elif [[ $A == *[?*]* ]]; then
		out=$out"$com ${A/%\*.\*/\*} $B"
	    else
		x=${A##*/} y=${B##*/}
		if [[ " txt bat sh baf bcs tp2 cre d bmp itm wav exe ogg tra bam sto 2da spl mus are dlg eff ids tis vvc wed mos pro wmp " == *\ `uncased "${x##*.}"`\ *
		    || $x == *.* && $y == *.* && ( ${x##*.} == *"${y##*.}"* || ${y##*.} == *"${x##*.}"* ) ]]; then
		    out=$out"$com $A $B"
		else
		    out=$out"x=($A/*) x=\${x%/\*}; $com \"\${x[@]}\" $B"
		fi
	    fi;;

	[Mm][Kk][Dd][Ii][Rr]|[Mm][Dd])
	    out=$out"mkdir -p $A";;

	[Rr][Mm][Dd][Ii][Rr]|[Rr][Dd])
	    if [[ $options == */[Ss]* ]]; then
		out=$out"rm -rf $A"
	    else
		out=$out"rmdir $A"
	    fi;;

	[Cc][Ll][Ss])
	    out=$out"clear 2>/dev/null || echo -en '\\e[H\\e[J'";;

	[Ee][Cc][Hh][Oo]*)
	    if [[ $A != [Oo][Ff][Ff] ]]; then
		x=${options%%[^$' \t']*}$whole_line
		if [[ $com == ????.* ]]; then
		    x=${com#????.}$x
		elif [[ $com == ???? ]]; then
		    x=${x:1}
		else
		    x=${com#????}$x
		fi
		out=$out"echo \""
		while [[ $x ]]; do
		    y=${x%%\$\{*}
		    x=${x:${#y}}
		    out=$out${y//\"/\\\"}
		    y=${x%%\}*}
		    out=$out$y
		    x=${x:${#y}}
		done
		out=$out\"
	    fi;;

	[Ee][Xx][Ii][Tt])
	    if [[ ! $A || $A == '${?}' ]]; then
		out=$out'exit `$Exit <<<${?}`'
	    elif [[ $A == 0 ]]; then
		out=$out'exit `$Exit <<<0`'
	    elif [[ $A == +([0-9]) ]]; then
		out=$out'exit `$Exit <<<'${A##*(0)}\`
	    else
		out=$out'exit `sed "s/[^0-9]*0*\([0-9][0-9]*\).*/\1/" <<<"'$A'" | $Exit`'
	    fi;;

	[Tt][Yy][Pp][Ee])
	    if [[ $A == [Nn][Uu][Ll] ]]; then
		out=$out"cat /dev/null"
	    elif [[ $redir == " >"* ]]; then
		out=$out"sed s/$'\r$'// $A"
	    else
		out=$out"tr -d '\0\377\376\r' <$A | sed -e 's/[]QZ]%/|/g' -e 's/[TWP]%/_/g'"
	    fi;;

	*[Ww][Ee][Ii][Dd][Uu]*)
	    x=$(dequote "$whole_line" all)
	    if [[ $x == *.[Tt][Pp]2* ]]; then
		y=${x#*.[Tt][Pp]2}
		z=" ${x:0:${#x}-${#y}}\\"
		while [[ $z == *\\ ]]; do
		    z=${z%[$' \t']*\\}
		done
		A=${x:${#z}:${#x}-${#y}-${#z}}
		out=$out"Install-mod.sh -batch -nogame $A; mv -f $A-mac ${A/%.[Tt][Pp]2/-mac.tp2}; \"\$WeiDU\"$z ${A/%.[Tt][Pp]2/-mac.tp2}$y"
	    else
		out=$out"\"\$WeiDU\" ${x//\*/\\*}"
	    fi;;

	*[Ss][Nn][Dd]2[Aa][Cc][Mm]*)
	    out=$out"snd2acm $A "$(dequote "$line" all);;

	[Nn][Oo][Tt][Ee][Pp][Aa][Dd]*|[Ee][Xx][Pp][Ll][Oo][Rr][Ee][Rr]*|[Vv][Ii][Ee][Ww])
	    if [[ $batch ]]; then
		if [[ $redir || $out == *"| " ]]; then
		    out=$out": "; else out=$out"# "
		fi
	    elif [[ ! -s /tmp/readme.$$ && $line != *\"\$* ]]; then
		echo $A > /tmp/readme.$$
		if [[ $redir || $out == *"| " ]]; then
		    out=$out": "; else out=$out"# "
	    fi  fi
	    out=$out"open $A "$(dequote "$line" all);;

	[Ss][Ee][Tt][Uu][Pp]?*)
	    if [[ ! $bwi ]]; then
		out=$out"$com "$(dequote "${whole_line//\*/\\*}" all)
	    else
		out=$out"Install-mod.sh -batch=\"$bwi "
		x=${whole_line%%*([$' \t'])} z=\'\"\$
		while [[ $x ]]; do
		    y=${x%%[$z]*}
		    out=$out$y
		    x=${x:${#y}}
		    if [[ ! $x ]]; then
			break
		    fi
		    if [[ $x == \'* ]]; then
			out=$out\\\'
			x=${x:1}
		    elif [[ $x == \$\{* ]]; then
			y=${x#*\}}
			out=$out${x:0:${#x}-${#y}}
			if [[ ${#y} == ${#x} ]]; then
			    break
			fi
			x=$y
		    elif [[ $x == \"* ]]; then
			if [[ $z == \'* ]]; then z=\"\$; else z=\'\"\$; fi
			x=${x:1}
			out=$out\\\"
		    else
			out=$out${x:0:1}
			x=${x:1}
		    fi
		done
		out=$out"\" $com"
	    fi;;

	[Cc][Aa][Ll][Ll])
	    last_com_was_call=1
	    if [[ $A == :* ]]; then
		out=$out"(Arg=(${A#:} "$(dequote "$line" all)"); "`tolower "${A/:/_}"`\)
	    else
		A=${A%.[Bb][Aa][Tt]}.bat B=$mod/${A//\\} x=${B%????}

		if [[ $B -nt $x.sh ]]; then
		    sed s/$'\r$'// "$B" | create_bash_script "${x##*/}" > "$x.sh"
		else
		    y=`grep -i "/${B##*/}$" "$mod/.batlist" | head -1`
		    if [[ ! $y || ! -e $y ]]; then
			out=$out"if [[ ! -f ${A%???}sh ]]; then Install-mod.sh -batch $A; fi; "
		    elif [[ $y -nt ${y%???}sh ]]; then
			x=${y##*/}
			sed s/$'\r$'// "$y" | create_bash_script "$x" > "${y%???}sh"
		fi  fi
		if [[ $sub_bat ]]; then
		    out=$out"(Arg=($A "$(dequote "$line" all)"); source ${A%???}sh)"
		else
		    out=$out"bash ${A%???}sh "$(dequote "$line" all)
	    fi  fi;;

	[Pp][Aa][Uu][Ss][Ee])
	    if [[ $batch && ! $bwi ]]; then
		if [[ $redir || $out == *"| " ]]; then
		    out=$out": "; else out=$out"# "
	    fi  fi
	    out=$out"read -p 'Press RETURN to continue...'";;

	[Vv][Ee][Rr])
	    out=$out"echo \"\"";;

	[Gg][Oo][Tt][Oo])
	    if [[ $A == ":"[Ee][Oo][Ff] ]]; then
		out=$out'exit `$Exit <<<${?}`'
	    else
		out=$out`tolower "_${A#:}"`
		if [[ $start_saving -ne 0 ]]; then
		    out=$out"; return"
	    fi	fi;;

	?(*/)?([Mm])[Tt][Ee][Ee]?(.[Ee][Xx][Ee]))
	    x=
	    if [[ $options == */[Aa+]* ]]; then
		x=" -a"
	    fi
	    out=$out"tee$x $A "$(dequote "$line" all);;

	[Pp][Aa][Tt][Hh]=*)
	    if [[ $redir || $out == *"| " ]]; then
		out=$out": "; else out=$out"# "; fi
	    out=$out\'$com\ ${whole_line//\'/\'\\\'\'}\';;

	*)
	    if [[ ${com##*/} == $extra_name || ${com##*/} == ${extra_name%.[Ee][Xx][Ee]} ]]; then
		out=$out"(exec -a '${com##*/}' \"\$WeiDU\" --noautoupdate --no-exit-pause ${whole_line//[*]/\\*})"
	    elif [[ $com == *\$\{* ]]; then	# Geez, Leonardo ...
		out=$out"(echo $com "$(dequote "$whole_line" all)'> /tmp/$$-bs.bat; Install-mod.sh -batch -nogame /tmp/$$-bs.bat; source /tmp/$$-bs.sh)'
	    elif [[ $com != *.* ]]; then
		if [[ $redir || $out == *"| " ]]; then
		    out=$out": "; else out=$out"# "; fi
		out=$out\'$com\ ${whole_line//\'/\'\\\'\'}\'
	    elif [[ $com == .* ]]; then
		if [[ $redir || $out == *"| " ]]; then
		    out=$out": "; else out=$out"# "; fi
		out=$out\'$com\ ${whole_line//\'/\'\\\'\'}\'
	    elif [[ $com == *.[Ee][Xx][Ee] ]]; then
		if [[ $redir || $out == *"| " ]]; then
		    out=$out": "; else out=$out"# "; fi
		out=$out\'$com\ ${whole_line//\'/\'\\\'\'}\'
	    else
		if [[ $batch ]]; then
		    if [[ $redir || $out == *"| " ]]; then
			out=$out": "; else out=$out"# "
		    fi
		elif [[ ! -s /tmp/readme.$$ && $com != *\"\$* ]]; then
		    echo "$com" > /tmp/readme.$$
		    if [[ $redir || $out == *"| " ]]; then
			out=$out": "; else out=$out"# "; fi
		fi
		out=$out"open $com"
	    fi;;
    esac

    if [[ $out == *"; " && ( ! $redir || $redir == "; "* ) ]]; then
	out=${out:0:${#out}-2}
    fi
    OUTPUT="$out$redir"
    redir=
}

modify_tp2_file () {
    local x q out y in_copy= inlined=

    sed s/$'\r$'// "$1" | tr \\0 \\200 | while read -r ; do
	if [[ $in_copy && $REPLY != *([$' \t'])[~\"]* ]]; then
	    in_copy=""
	fi
	if [[ $inlined ]]; then
	    echo "$REPLY"
	    if [[ $REPLY == *">>>>>>>>"* ]]; then
		if [[ $inlined == *.[Bb][Aa][Tt] ]]; then
		    x=${inlined//[^.a-zA-Z0-9]/_}
		    echo "COPY ~$inlined~ ~/tmp/$x~"
		    echo "AT_NOW ~Install-mod.sh -batch /tmp/$x; mv -f /tmp/${x%.???}.sh \"${inlined%.???}.sh\"; rm -f /tmp/$x~"
		fi
		inlined=
	    fi
	elif [[ $in_copy || $REPLY == *([$' \t'])COPY\ * ]]; then
	    REPLY=${REPLY//.[bB][aA][tT]/.sh}
	    x=${REPLY%%//*}; x=${x%[~\"]*}; x=${x##*[~\"]}
	    if [[ $x && $x != *[/\\]* && ! -d "$x" ]]; then
		for ((i=0; $i < ${#added_files[*]}; i++)); do
		    if [[ $x == `uncased "${added_files[$i]}" no` ]]; then
			x=
			break
		    fi
		done
		if [[ $x ]]; then
		    added_files[${#added_files[*]}]=$x
		fi
		in_copy=true
	    fi
	    echo "$REPLY"
	elif [[ $REPLY == *([$' \t'])AT_* ]]; then
	    # throw away all context changes from select_batch_command
	    local cdpat= deferred= redir= save_til_end= start_saving= OUTPUT
	    x=${REPLY##*([$' \t'])AT_*([^$' \t'])*([$' \t'])}
	    q=${x:0:1}
	    x=${x:1}; y=${x#*$q}; x=${x%%$q*} x=${x##*([$' \t'])}
	    if [[ $x == sh\ * || $x == bash\ * ]]; then
		OUTPUT=${x/sh/sh -O nocaseglob}
	    elif [[ $x == rm\ * || $x == cp\ * || $x == mv\ * ]]; then
		OUTPUT="shopt -s nocaseglob; $x"
	    else
		select_batch_command "$x" dont_replace init </dev/null
		OUTPUT=${OUTPUT/#\#[$' \t']/: }
	    fi
	    echo "${REPLY%%$q*}$q${OUTPUT%%*([$' \t'])}$q$y"
	elif [[ $REPLY == *([$' \t'])README* ]]; then
	    x=${REPLY##*([$' \t'])README*([$' \t'])}
	    q=${x:0:1}
	    x=${x:1}; y=${x#*$q}; x=${x%%$q*}
	    echo "$x" > /tmp/readme.$$
	    echo "// $REPLY"
	elif [[ $REPLY == *([$' \t'])BACKUP* && (! $batch || $batch == true) ]]; then
	    x=${REPLY##*([$' \t'])BACKUP*([$' \t'])}
	    q=${x:0:1}
	    if [[ $q != [\"\~] ]]; then
		echo "${REPLY:0:${#REPLY}-${#x}}~${BGdir//\//_}~"
	    else
		echo "${REPLY%$q*}${BGdir//\//_}$q"
	    fi
	elif [[ $REPLY == *([$' \t'])"<<<<<<<<"* ]]; then
	    inlined=${REPLY##*([$' \t'])<<<<<<<<*([$' \t'])}
	    echo "$REPLY"
	else
	    echo "$REPLY"
	fi
    done
}

load_mod_dir () {
    local mod x z failed= orig_dir=$PWD

    if [[ -d $1 && $1 != *.pkg ]]; then
	mod=$1
	cd "$mod"
    else
	if [[ $2 ]]; then
	    mod=$PWD
	else
	    mod=${1##*/} mod=$BGdir/BWPmods/${mod%.*}
	    rm -rf "$mod" >& /dev/null
	    mkdir -p "$mod" 2>/dev/null
	    cd "$mod"
	fi

	x=`file "$1"`
	if [[ $1 == *.dmg ]]; then
	    ls -1d /Volumes/* > /tmp/pkg.$$
	    open -W "$1"
	    z=`ls -1d /Volumes/* | comm -13 /tmp/pkg.$$ -`
	    if [[ $z ]]; then
		cp -rp "$z"/* "$mod"
		diskutil eject "$z" > /dev/null
	    else
		failed=1
	    fi
	    rm -f /tmp/pkg.$$
	elif [[ $1 == *.pkg ]]; then
	    if unar "$1"/Contents/Archive.pax* 2>&1 | grep -v '\.\.\.$\|OK\.$' >/dev/null; then
		failed=1
	    elif [[ ! -e Archive.pax ]]; then
		: finished with this
	    elif unar Archive.pax 2>&1 | grep -v '\.\.\.$\|OK\.$' >/dev/null; then
		failed=1
	    else
		rm -f Archive.pax
	    fi
	elif [[ $1 == *.[Rr][Aa][Rr] || $x == *\ [Rr][Aa][Rr]\ * ]]; then
	    unrar x -y "$1" >& /dev/null
	    failed=$?
	elif [[ $1 == *.7[Zz] || $x == *\ 7-[Zz][Ii][Pp]\ * ]]; then
	    "$tooldir"/7z x -y "$1" >& /dev/null
	    failed=$?
	elif [[ $1 == *.[Zz][Ii][Pp] || $x == *\ [Zz][Ii][Pp]\ * ]]; then
	    unzip -o "$1" >& /dev/null
	    failed=$?
	else
	    mkdir .unar
	    cd .unar
	    if unar "$1" 2>&1 | grep -v '\.\.\.$\|OK\.$' >/dev/null; then
		cd ..
		rm -rf .unar
		failed=1
		if "$tooldir"/7z x -y "$1" >& /dev/null; then
		    rm -f UPX[0-9]*
		    for z in \$*; do
			mv -f "$z"/* . >& /dev/null && rm -rf "$z"
		    done
		    failed=0
		fi
	    else
		failed=0
		mv -f * ..
		cd ..
		rmdir .unar
	fi  fi

	if [[ $failed -ne 0 ]]; then
	    echo "| $1 failed to extract; skipping" 1>&2
	    if [[ ! $2 ]]; then
		cd "$orig_dir"
		rm -rf "$mod"
	    fi
	    return
	fi

	if [[ $2 ]]; then
	    echo "$mod"
	    return
    fi  fi

    z=(*)
    if [[ $z == \* ]]; then
	echo "| ${mod##*/} is an empty directory; skipping" 1>&2
	cd "$orig_dir"
	if [[ ! -d $1 ]]; then
	    rm -rf "$mod"
	fi
	return
    fi
    if [[ ${#z[@]} -eq 1 && -d $z && $z != *.pkg ]]; then
	if [[ -e $z/$z ]]; then
	    mv -f "$z" ".$z"
	    z=.$z
	fi
	if [[ `ls "$z"/Setup*.exe 2>/dev/null` ]]; then
	    mv -f "$z"/* .
	    rm -rf "$z"
	else
	    cp -fpr "$z"/* .
    fi  fi

    if [[ `ls *.pkg 2>/dev/null` || ! `ls -d */. 2>/dev/null` ]]; then
	for z in *.{pkg,zip,7z,rar,exe}; do
	    if [[ $z == *[^*].pkg || `file "$z" | fgrep -i " archive"` ]]; then
		if [[ $1 == "$mod" ]]; then
		    mod=${1##*/} mod=$BGdir/BWPmods/${mod%.*}
		    rm -rf "$mod" >& /dev/null
		    mkdir -p "$mod" 2>/dev/null
		    cp -fpr "$1"/* "$mod"
		    cd "$mod"
		fi
		mv -f "$z" ".$z"
		if [[ ! `load_mod_dir ".$z" same` ]]; then
		    cd "$orig_dir"
		    rm -rf "$mod"
		    return
		fi
		rm -rf ".$z"
	    fi
	done
    fi

    cd "$orig_dir"
    echo "$mod"
}

#### script starts here ####

export LANG=C
orig_dir=`pwd -P`

if [[ $0 != /* ]]; then
    tooldir=$orig_dir/$0
else
    tooldir=$0
fi
tooldir=${tooldir%/*}
export PATH="/bin:/usr/bin:$tooldir:$PATH"

weidu=weidu-mac

while [[ $1 == -* ]]; do
    if [[ $1 == -batch* ]]; then
	x=${1#-batch}
	if [[ $x == =* ]]; then
	    batch=${x#=}
	else
	    batch=true
	fi
    elif [[ $1 == -bwi* ]]; then
	x=${1#-bwi}
	if [[ $x == =* ]]; then
	    bwi=${x#=}
	else
	    bwi=true
	fi
    elif [[ $1 == -nogame ]]; then
	nogame=true
    fi
    shift
done

if [[ $nogame ]]; then
    BGdir=$orig_dir
    WeiDU=$weidu
else
    if [[ $BGdir && -f $BGdir/Chitin.key ]]; then
	cd "$BGdir"
    elif [[ ! -f Chitin.key ]]; then
	if [[ -f ../Chitin.key ]]; then
	    cd ..
	else
	    if [[ ! `whereis osascript` ]]; then
		echo "| Enter your Baldur's Gate directory"
		echo -n "| => "
		read REPLY
	    else
		REPLY=`osascript -e 'tell application "Terminal" to return (POSIX path of (choose folder with prompt "Choose your Baldurs Gate folder"))'`
	    fi

	    cd "${REPLY%%*(/)*( )}"
	    if [[ ! -f Chitin.key ]]; then
		echo "| WARNING: $PWD doesn't look like a Baldur's Gate folder."
		## exit 1
    fi  fi  fi

    BGdir=`pwd -P`
    export PATH="$BGdir:$PATH"

    if [[ ! $WeiDU ]]; then
	x=$PATH
	while [[ $x ]]; do		# 10.4 has a broken 'which' command
	    y=${x%%:*}
	    if [[ -x "$y/$weidu" && -f "$y/$weidu" ]]; then
		z=$("$y/$weidu" --help </dev/null 2>&1 | egrep -i 'weidu.*version' | head -1 | sed -e 's/.*]//')
		z=${z##*([^0-9])}
		z=${z%%*([^0-9])}
		if [[ $z -gt $best ]]; then
		    best=$z
		    WeiDU=$y/$weidu
	    fi  fi
	    x=${x:${#y}+1}
	done

	if [[ ! $WeiDU ]]; then
	    echo "| Can't find weidu-mac; quitting"
	    exit 1
    fi  fi

    if [[ $WeiDU != $BGdir/weidu-mac ]]; then
	rm -f "$BGdir/weidu-mac"
	cp -f "$WeiDU" "$BGdir" 2>/dev/null
    fi

    if [[ ! $batch ]]; then
	echo Using \"$WeiDU\" WeiDU version $best
	echo BGdir \"$BGdir\"
    fi

    if [[ ! -d override-orig ]]; then
	cp -rpf override override-orig 
	cp -pf "$tooldir"/BG2-missing/* override
    fi

    if [[ ! -f Chitin.key-orig ]]; then
	cp -pf Chitin.key Chitin.key-orig
    fi

    if [[ ! -f dialog.tlk-orig ]]; then
	cp -pf dialog.tlk dialog.tlk-orig
    fi
fi

export BGdir WeiDU

while [[ ! $quit_when_done ]]; do
    if [[ $1 ]]; then
	mod=$1
	shift
	if [[ ! $1 ]]; then
	    quit_when_done=true
	fi
    else
	echo ""
	echo "| Drop the Mod here and press RETURN (or type quit to exit)"
	echo -n "| => "
	read
	mod=${REPLY%%*(/)*( )}
	if [[ "$mod" == [qQ] || "$mod" == [eEqQ][xXuU][iI][tT] ]]; then
	    break
    fi	fi

    if [[ ! "$mod" ]]; then
	continue
    fi

    if [[ $mod != /* ]]; then
	mod=$orig_dir/$mod
    fi

    if [[ ! -e $mod ]]; then
	echo "| \"$mod\" does not exist."
	continue
    fi

    if [[ $mod == *.[Bb][Aa][Tt] ]]; then
	x=$mod
	if [[ $mod != */* ]]; then
	    mod=.
	else
	    mod=${mod%/*}
	fi
	if [[ ! -e $mod/.batlist ]]; then
	    find -H "$mod" -iname \*.bat -print > "$mod/.batlist"
	fi
	sed s/$'\r$'// "$x" | create_bash_script > "${x/%[Bb][Aa][Tt]/sh}"
	continue
    fi

    if [[ $mod == *.[Tt][Pp]2 ]]; then
	x=$mod
	if [[ $mod != */* ]]; then
	    mod=.
	else
	    mod=${mod%/*}
	fi
	if [[ ! -e $mod/.batlist ]]; then
	    find -H "$mod" -iname \*.bat -print > "$mod/.batlist"
	fi
	modify_tp2_file "$x" | tr \\200 \\0 > "$x-mac"
	continue
    fi

    setup=
    if [[ $mod == */[Ss][Ee][Tt][Uu][Pp]*[Ee][Xx][Ee] ]]; then
	if [[ -h $mod ]]; then
	    z=`ls -ld "$mod"` z=${z#* -> }
	    if [[ $z != /* ]]; then
		mod=${mod%/*}/$z
	    else
		mod=$z
	fi  fi

	setup=${mod##*/}
	mod=${mod%/*}

	if [[ $mod == */BWPmods ]]; then
	    z=${setup%.[Ee][Xx][Ee]} z=${z#[Ss][Ee][Tt][Uu][Pp]?}
	    rm -f "$mod/$z-dir"
	    mkdir "$mod/$z-dir" 2>/dev/null
	    if [[ -e $mod/${setup%.[Ee][Xx][Ee]}.tp2 ]]; then
		ln -s "$mod/${setup%.[Ee][Xx][Ee]}.tp2" "$mod/$z-dir" 
	    elif [[ -e $mod/$z.tp2 ]]; then
		ln -s "$mod/$z.tp2" "$mod/$z-dir" 
	    fi
	    touch "$mod/$z-dir/$setup"
	    if [[ -d $mod/$z ]]; then
		ln -s "../$z" "$mod/$z-dir"
	    else
		mkdir "$mod/$z-dir/$z" 2>/dev/null
	    fi
	    mod=$mod/$z-dir
	fi
    else
	mod=`load_mod_dir "$mod"`
	if [[ ! $mod ]]; then
	    continue
    fi  fi

    if (( `ls -di "$mod/." "$BGdir" | sort -un | wc -l` == 1 )); then
	echo "| Mod folder can't be the same as the Baldur's Gate folder"
	continue
    fi

    x="${mod//[ ][ ]/_}"
    x="${x//[\]\'\\\\\[\>\"\&\`\?\*\|\$\<\(\)]/_}" # remove bad chars in mod dir
    if [[ "$x" != "$mod" ]]; then
	x="${x##*/}"
	if [[ -e "$x" && ! -h "$x" ]]; then
	    echo "| Moving $x to $x-old"
	    mv -f "$x" "$x-old"
	fi
	rm -f "$x"
	ln -sf "$mod" "$x"
	mod="$x"
    else
  	mod="${mod#$BGdir/}"   # make mod dir relative if subdir of bgdir
    fi

    extra_arg=()
    if [[ $mod == */[Bb][Gg][Tt]*[Ii][Nn][Ss][Tt][Aa][Ll][Ll] ]]; then
	if [[ $batch != *--args-list*([\ s])p* ]]; then
	    while ((1)); do
		if [[ ! `whereis osascript` ]]; then
		    echo "| Enter your Tales of the Sword Coast folder here"
		    read -p "| => "
		else
		    REPLY=`osascript -e 'tell application "Terminal" to return (POSIX path of (choose folder with prompt "Choose your TotSC folder"))'`
		fi
		x=${REPLY%%*(/)*( )}
		if [[ -f $x/Chitin.key ]]; then
		    break
		fi
		x=$x"/Game Data"
		if [[ -f $x/Chitin.key ]]; then
		    break
		fi
		echo "| That doesn't look like a TotSC folder; try again"
	    done

	    touch "$x"/BGMain2.exe
	    extra_arg=(--args-list p "$x")
	    if [[ ! -e "Baldur's Gate II Lev 1" ]]; then
		makeLev1.sh <<<"$BGdir/Baldur's Gate II" > /dev/null
	    fi
	    if [[ ! -e BGMain.exe ]]; then
		dd if=/dev/zero of=BGMain.exe bs=4096 oseek=2000 count=1 2>/dev/null
	fi  fi

	if [[ ! -e "$BGdir/Baldur's Gate II Lev 1" ]]; then
	    makeLev1.sh <<<"$BGdir/Baldur's Gate II" > /dev/null
	    echo "| Use \"Baldur's Gate II Lev 1\" to start a BG1 game at Candlekeep."
	fi
	# There may be too many files to copy in over.bak
	sed -i "" '/^call BGT\\Install\\deldir override/,/^xcopy/s/^/:: /' "$mod/BGT/Install/Uninstall.bat"
    fi

    find -H "$mod" -iname \*.sh -exec chmod a+rx {} \; -exec sed -i "" -e s,\[^$' \t'\]\*/sox,sox,g -e s,\[^$' \t'\]\*/tisunpack,tisunpack,g -e s,$'\r$',, -e s,\[^\ \]*/\[Oo\]\[Vv\]\[Ee\]\[Rr\]\[Rr\]\[Ii\]\[Dd\]\[Ee\],\"\$BGdir\"/override,g {} \;

    ls "$mod"/*.[Tt][Pp]2 "$mod"/*/*.[Tt][Pp]2 > /tmp/tp2.$$ 2>/dev/null
    (touch `ls "$mod"/[Ss][Ee][Tt][Uu][Pp]*.command | sed s/command/exe/`) 2>/dev/null
    ls "$mod"/[Ss][Ee][Tt][Uu][Pp]*.[Ee][Xx][Ee] > /tmp/exe.$$ 2>/dev/null

    if [[ -s /tmp/exe.$$ ]]; then x=/tmp/exe.$$; else x=/tmp/tp2.$$; fi
    sed -e 's=.*/\(.*\)\.[TteE][PpxX][2eE]$=-\1.exe=' -e 's=-[Ss][Ee][Tt][Uu][Pp]==' -e 's=^=Setup=' $x | sort -fu -o /tmp/exe.$$

    if [[ ! -e $mod/.batlist ]]; then
	ls "$mod"/*.bat 2>/dev/null > "$mod/.batlist"
	if [[ `ls -1d "$mod"/*/. 2>/dev/null` ]]; then
	    find -H "$mod"/*/. -iname \*.bat -print | sed s,/\\./,/, >> "$mod/.batlist"
    fi  fi

    added_files=() extra_name=
    IFS=$'\n'; for tp2 in `cat /tmp/tp2.$$`; do IFS=$' \t\n'
	x=${mod%/*}/
	if [[ $tp2 -nt $tp2-mac ]]; then
	    echo "TP2 file ${tp2:${#x}}"
	    modify_tp2_file "$tp2" | tr \\200 \\0 > "$tp2-mac"
	fi
	x=${tp2##*/}
	if [[ $x == [Ss][Ee][Tt][Uu][Pp]* ]]; then
	    rm -f "$x"
	    cp "$tp2-mac" "Setup-${x#[Ss][Ee][Tt][Uu][Pp]?}"
	fi
    done
    rm -f /tmp/tp2.$$

    if [[ ! -s $mod/.batlist && ! -s /tmp/exe.$$ ]]; then
	echo "| This doesn't seem to be a Mod folder; no .TP2 or .BAT files; skipping"
	rm -f "$mod/.batlist" /tmp/exe.$$
	continue
    fi

    IFS=$'\n'; for bat in `cat "$mod/.batlist"`; do IFS=$' \t\n'
	x=${bat%.???}
	echo "Batch file ${bat##*/}"
	if [[ $bat -nt $x.sh ]]; then
	    sed s/$'\r$'// "$bat" | create_bash_script "${x##*/}" > "$x.sh"
	fi
    done

    IFS=$'\n'; for exe in `cat /tmp/exe.$$`; do IFS=$' \t\n'
	x=${exe%.[Ee][Xx][Ee]}
	rm -f "$x.exe" "$x.tp2"
	tp2=$mod/$x.tp2
	if [[ ! -e $tp2 ]]; then
	    r=${x#[Ss][Ee][Tt][Uu][Pp]?}
	    tp2=$mod/$r.tp2
	    if [[ ! -e $tp2 ]]; then
		tp2=$mod/$r/$x.tp2
		if [[ ! -e $tp2 ]]; then
		    tp2=$mod/$r/$r.tp2
		    if [[ ! -e $tp2 ]]; then
			echo "| Can't find matching .exe and .tp2 files for ${exe%.[Ee][Xx][Ee]}"
			continue
	fi  fi  fi  fi
	# ln -fs "$WeiDU" "$x.exe"
	cp -f "$tooldir/rep_weidu" "$x.exe"
	cp -f "$tp2-mac" "${tp2##*/}"
    done

    to_be_removed=()
    for file in "$mod"/* ; do
	x=${file##*/}
	if [[ -f $file && $x != *.* || $x == *.[Tt][Pp]2* || $x == *.[Ee][Xx][Ee]* || $x == *.command || $x == *[Dd][Ee][Bb][Uu][Gg]* || $x == *-save || $x == *.rej || $x == *.orig || $x == *-orig || $x == *.[Tt][Xx][Tt] ]]; then
	    continue
	fi
	if [[ -h $x ]]; then
	    rm -f "$x"
	elif [[ -f $x ]]; then
	    echo "| Moving old $x to $x-old"
	    mv -f "$x" "$x-old"
	elif [[ -d $x ]]; then
	    if [[ ! -d $file ]]; then
		continue
	    fi
	    echo "WARNING: This mod dumps files into $x, so it may be uninstallable."
	    for df in "$file"/* ; do
		y=${df##*/}
		if [[ -e $x/$y && `cmp -s "$x/$y" "$df"` ]]; then
		    echo "| Moving old $x/$y to $x-old/$y"
		    mkdir "$x-old" 2>/dev/null
		    if [[ ! -e $x-old/$y ]]; then
			mv -f "$x/$y" "$x-old/$y"
		fi  fi
		cp -Rf "$df" "$x/$y"
	    done
	    continue
	fi
	ln -fs "$file" "$x"
	if [[ ! -d $file && $x != *.sh ]]; then
	    to_be_removed[${#to_be_removed[*]}]=$x
	fi
    done

    if [[ ! $batch ]]; then
	if [[ -s /tmp/readme.$$ ]]; then
	    x="$(< /tmp/readme.$$)"
	    open "${x//%[lL][aA][nN][gG][uU][aA][gG][eE]%/English}"
	fi

	if [[ ${#added_files[*]} != 0 ]]; then
	    echo "| These files will be added to Baldur's Gate folder:"
	    echo "| ${added_files[*]}"
	fi
	echo "|"
	echo "| Fixing broken files in mod:"
    fi
    rm -f /tmp/readme.$$

    if [[ ! -e $mod/Check.debug ]]; then
	chmod -R -L 775 "$mod"
	"$WeiDU" --noautoupdate --no-exit-pause --log "$mod"/Check.debug "$tooldir"/check.tp2 <<<"$mod" >& /dev/null # | egrep '^(\(|ERROR)' | sed 's,.*/\([^)]*\)),\1:,'
	find "$mod" -iname \*.sh -exec sed -i "" s/$'\r'// {} \;
    fi

    if [[ ! -s /tmp/exe.$$ ]]; then
	ls *.sh 2>/dev/null | grep -v -i uninst > /tmp/instbat.$$
	if [[ `wc -l /tmp/instbat.$$` -ne 1 ]]; then
	    echo "|"
	    echo "| Mod converted for Macs.  Read the Mod's ReadMe to find out what to do next."
	    continue
	fi
	if [[ ! $batch ]]; then
	    echo -n "| Mod converted for Macs.  Run `cat /tmp/instbat.$$`? (Yes/no): "
	    read x
	    if [[ $x == [Nn]* ]]; then
		echo "| OK. When you're ready, run `cat /tmp/instbat.$$`"
		if [[ ${#to_be_removed[*]} != 0 ]]; then
		    echo "| You can then remove: ${to_be_removed[*]}"
		fi
		continue
	fi  fi
	bash "`/tmp/instbat.$$`"
	rm /tmp/instbat.$$
    else
	if [[ ! $batch ]]; then
	    echo -n "| Mod converted for Macs.  Run WeiDU to install it now? (Yes/no): "
	    read x
	    if [[ $x == [nN]* ]]; then
		echo "| OK. When you're ready, run: " $(< /tmp/exe.$$) "${extra_arg[@]}"
		if [[ ${#to_be_removed[*]} != 0 ]]; then
		    echo "| You can then remove: ${to_be_removed[*]}"
		fi
		continue
	fi  fi

	IFS=$'\n'; for exe in `cat /tmp/exe.$$`; do IFS=$' \t\n'
	    if [[ $setup && $setup != `uncased "$exe"` ]]; then
		#echo "$setup: skipping $exe"
		continue
	    fi
	    if [[ ! $batch ]]; then
		echo Running "$exe" ...
		"$exe" "${extra_arg[@]}"
	    elif [[ $batch == true ]]; then
		"$exe" "${extra_arg[@]}"
	    elif [[ $batch != manual* ]]; then
		batch=${batch#* }
		if [[ $batch != *--force-install* ]]; then
		    batch=$batch" --yes"
		fi
		eval y=($batch)
		"$exe" "${y[@]}"
	    else
		batch=${batch#* }
		if [[ $batch != *--force-install* ]]; then
		    z="$batch --yes"
		    echo "$exe: here BWP installs all components" 1>&2
		else
		    z=${batch%--force-install*}
		    y=${batch#*--force-install* }
		    if [[ $y == *--* ]]; then
			z=$z${y/*--/--}
			y=${y%%--*}
		    fi
		    echo "$exe: here BWP installs component# $y" 1>&2
		fi
		read -p "${exe//?/ }  Accept this (type Yes) or run WeiDU manually? " x 1>&2 < /dev/tty
		if [[ $x == [Yy]* ]]; then
		    eval x=(`echo $batch`)
		    "$exe" "${x[@]}"
		else
		    eval x=(`echo $z`)
		    "$exe" "${x[@]}"
	    fi  fi
	    echo "|______________________________________________________________________________"
	    echo "|"
	done
    fi

    rm -f "${to_be_removed[@]}"
    rm -f /tmp/*.$$ override/*.d override/*.baf

    if [[ -e override/songlist.2da ]]; then
	sed -i "" '/^[1-9][0-9][0-9]/d' override/songlist.2da
    fi
done

if [[ ! $batch ]]; then
    echo Thanks for playing
fi
exit 0
